/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "houdini.pmh"

#define FE_STH_DEBUG	FALSE

namespace fe
{

SurfaceTrianglesHoudini::SurfaceTrianglesHoudini(void):
	m_subIndex(-1)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini()\n");
#endif
}

SurfaceTrianglesHoudini::~SurfaceTrianglesHoudini(void)
{
}

Protectable* SurfaceTrianglesHoudini::clone(Protectable* pInstance)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::clone\n");
#endif

	SurfaceTrianglesHoudini* pSurfaceTrianglesHoudini= pInstance?
			dynamic_cast<SurfaceTrianglesHoudini*>(pInstance):
			new SurfaceTrianglesHoudini();

	SurfaceTriangles::clone(pSurfaceTrianglesHoudini);

	// TODO copy attributes

	return pSurfaceTrianglesHoudini;
}

void SurfaceTrianglesHoudini::resizeArrays(U32 a_count)
{
	m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
			a_count*sizeof(SpatialVector));
	m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
			a_count*sizeof(SpatialVector));
}

void SurfaceTrianglesHoudini::cache(void)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::cache\n");
#endif

	//* NOTE not doing any Record serialization

	//* convert Houdini data to vertex and normal arrays

	const I32 entries=m_pGdp->primitives().entries();
	if(m_subIndex>=entries)
	{
		feLog("SurfaceTrianglesHoudini::cache excessive subIndex %d/%d\n",
				m_subIndex,entries);
		return;
	}
	const U32 count=(m_subIndex>=0)? 1: entries;

	m_vertices=0;
	U32 allocated=0;

	GEO_AttributeHandle handlePoint=
			m_pGdp->getAttribute(GEO_POINT_DICT,GEO_STD_ATTRIB_NORMAL);

	U32 total=0;
	for(U32 m=0;m<count;m++)
	{
		const U32 index=(m_subIndex>=0)? m_subIndex: m;
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(index);

		//* see GEO_PrimType.h
		const U32 primitiveID=pPrimitive->getPrimitiveId();
#if FE_STH_DEBUG
		feLog("index %d primitiveID %p\n",index,primitiveID);
#endif

		if(primitiveID==GEOPRIMMESH)
		{
			//* NOTE castTo doesn't work, but RTTI does
//			const GEO_PrimMesh* pPrimMesh=
//					(const GEO_PrimMesh*)pPrimitive->castTo();
			const GEO_PrimMesh* pPrimMesh=
					dynamic_cast<const GEO_PrimMesh*>(pPrimitive);

			const U32 columns=pPrimMesh->getNumCols();
			const U32 rows=pPrimMesh->getNumRows();
			const BWORD wrapU=pPrimMesh->isWrappedU();
			const BWORD wrapV=pPrimMesh->isWrappedV();
#if FE_STH_DEBUG
			feLog("rows %d columns %d wrapped %d %d\n",
					rows,columns,wrapU,wrapV);
#endif

			if(!columns || !rows)
			{
				feLog("SurfaceTrianglesHoudini::cache as GEO_PrimMesh"
						" primitive %d/%d columns=%d rows=%d (degenerate)\n",
						index,count,columns,rows);
				continue;
			}

			U32 countU=columns+wrapU-1;
			U32 countV=rows+wrapV-1;

			m_vertices+=6*countU*countV;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
				resizeArrays(desired);
				allocated=desired;
			}

			SpatialVector vert0[columns];	// first
			SpatialVector vert1[columns];	// previous
			SpatialVector vert2[columns];	// current
			SpatialVector norm0[columns];
			SpatialVector norm1[columns];
			SpatialVector norm2[columns];
			for(U32 v=0;v<rows+wrapV;v++)
			{
				if(v)
				{
					for(U32 u=0;u<columns;u++)
					{
						vert1[u]=vert2[u];
						norm1[u]=norm2[u];
					}
				}
				if(v<rows)
				{
					for(U32 u=0;u<columns;u++)
					{
						//* location
						UT_Vector4 utVector4=
								pPrimMesh->getVertex(v,u).getPt()->getPos();
						vert2[u]=SpatialVector(utVector4.data());

						//* normal
						const U32 pointIndex=pPrimMesh->getVertex(v,u)
								.getBasePt()->getNum();
						const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
						handlePoint.setElement(pPoint);
						utVector4=handlePoint.getV4(0);
						norm2[u]=SpatialVector(utVector4.data());

#if FE_STH_DEBUG
						feLog(" %d %d %s %s\n",u,v,
								c_print(vert2[u]),c_print(norm2[u]));
#endif
					}
				}
				else
				{
					for(U32 u=0;u<columns;u++)
					{
						vert2[u]=vert0[u];
						norm2[u]=norm0[u];
					}
				}
				if(v)
				{
					U32 u;
					for(u=0;u<columns-1;u++)
					{
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
						//* lower right
						m_pVertexArray[total]=vert1[u+1];
						m_pNormalArray[total]=norm1[u+1];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[u+1];
						m_pNormalArray[total]=norm2[u+1];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[u+1];
						m_pNormalArray[total]=norm2[u+1];
						total++;
						//* upper left
						m_pVertexArray[total]=vert2[u];
						m_pNormalArray[total]=norm2[u];
						total++;
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
					}
					//* NOTE u=columns-1
					if(wrapU)
					{
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
						//* lower right
						m_pVertexArray[total]=vert1[0];
						m_pNormalArray[total]=norm1[0];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[0];
						m_pNormalArray[total]=norm2[0];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[0];
						m_pNormalArray[total]=norm2[0];
						total++;
						//* upper left
						m_pVertexArray[total]=vert2[u];
						m_pNormalArray[total]=norm2[u];
						total++;
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
					}
				}
				else
				{
					for(U32 u=0;u<columns;u++)
					{
						vert0[u]=vert2[u];
						norm0[u]=norm2[u];
					}
				}
			}
		}
		if(primitiveID==GEOPRIMPOLY)
		{
			const U32 subCount=pPrimitive->getVertexCount();
			if(subCount!=3)
			{
				feLog("SurfaceTrianglesHoudini::cache as GEO_PrimPoly"
						" primitive %d/%d subCount=%d (not triangles?)\n",
						index,count,subCount);
				continue;
			}

			m_vertices+=3;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
				resizeArrays(desired);
				allocated=desired;
			}

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				//* location
				UT_Vector4 utVector4=
						pPrimitive->getVertex(subIndex).getPt()->getPos();
				m_pVertexArray[total]=SpatialVector(utVector4.data());

#if FE_STH_DEBUG
				feLog("poly prim %d/%d sub %d/%d %s\n",index,count,
						subIndex,subCount,c_print(m_pVertexArray[total]));
#endif

				//* normal
				const U32 pointIndex=
						pPrimitive->getVertex(subIndex).getBasePt()->getNum();
				const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
				handlePoint.setElement(pPoint);
				utVector4=handlePoint.getV4(0);
				m_pNormalArray[total]=SpatialVector(utVector4.data());

				total++;
			}
		}
	}
	if(allocated!=m_vertices)
	{
		resizeArrays(m_vertices);
		allocated=m_vertices;
	}

	if(total!=m_vertices)
	{
		feLog("SurfaceTrianglesHoudini::cache converted %d/%d vertices\n",
				total,m_vertices);
		for(U32 m=total;m<m_vertices;m++)
		{
			set(m_pVertexArray[m]);
			set(m_pNormalArray[m],0.0,1.0,0.0);
		}
	}

#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::cache %d vertices\n",m_vertices);
#endif

	calcBoundingSphere();
}

void SurfaceTrianglesHoudini::calcBoundingSphere(void)
{
	if(!m_vertices)
	{
		return;
	}

	m_center=m_pVertexArray[0];
	for(U32 m=1;m<m_vertices;m++)
	{
		m_center+=m_pVertexArray[m];
	}
	m_center*=1.0/m_vertices;

	Real max2=0.0;
	for(U32 m=1;m<m_vertices;m++)
	{
		SpatialVector radial=m_pVertexArray[m]-m_center;
		Real dist2=magnitudeSquared(radial);
		if(max2<dist2)
		{
			max2=dist2;
		}
	}
	m_radius=sqrt(max2);

#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::calcBoundingSphere"
			" vertices %d center %s radius %.6G\n",
			m_vertices,c_print(m_center),m_radius);
#endif
}

} // namespace
