/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OBLD_DEBUG			FALSE

#include "grass/grass.pmh"

using namespace fe;

void BladeOp::initialize(void)
{
	catalog<Real>("Tension")=1e4;
	catalog<Real>("Tension","high")=1e5;
	catalog<String>("Tension","hint")="Angular springiness.";

	catalog<Real>("Drag")=1e2;
	catalog<Real>("Drag","high")=1e3;
	catalog<String>("Drag","hint")="Angular sluggishness.";

	catalog<Real>("Bend")=1e-2;
	catalog<Real>("Bend","high")=1e0;
	catalog<String>("Bend","hint")=
			"Flatness curvature.  This is added to the rest position.";

	catalog<Real>("Mass")=1e0;
	catalog<Real>("Mass","high")=1e1;
	catalog<String>("Mass","hint")="Mass per vertex point.";

	catalog<Real>("Wind")=0;
	catalog<Real>("Wind","high")=1e3;

	catalog<Real>("Turbulence")=1e2;
	catalog<Real>("Turbulence","high")=1e3;

	catalog<Real>("Grab Bias")=0.0;
	catalog<Real>("Grab Bias","min")= -1.0;
	catalog<Real>("Grab Bias","max")=1.0;
	catalog<String>("Grab Bias","hint")=
			"Tune how easily the blade tips stick to the collider.";

	catalog<Real>("TimeStep")=1.0/60.0;
	catalog<Real>("TimeStep","high")=1e-1;

	catalog<Real>("Restoration")=0.0;
	catalog<Real>("Restoration","high")=1e5;

	catalog<Real>("Collision Gap")=0.0;
	catalog<String>("Collision Gap","hint")=
			"Impose extra distance from the collider.";

	catalog<I32>("Collision Method")=0;
	catalog<I32>("Collision Method","high")=3;
	catalog<String>("Collision Method","hint")=
			"0 = bounding sphere\n"
			"1 = nearest point\n"
			"2 = ray cast (one step)\n"
			"3 = ray cast (halvsies)";

	catalog<I32>("Refinement")=0;
	catalog<I32>("Refinement","high")=4;
	catalog<I32>("Refinement","max")=8;
	catalog<String>("Refinement","hint")=
			"Maximum level of collider subdivision."
			"  Refinement may be adaptive and non-uniform.";

	catalog<String>("Tangent Attribute")="tangent";
	catalog<String>("Tangent Attribute","hint")=
			"Name of vector attribute on input curves"
			" describing the direction along which rotations can occur.";

	catalog<String>("Facing Attribute")="facing";
	catalog<String>("Facing Attribute","hint")=
			"Name of vector attribute on input curves"
			" describing the facing side.  This may get rotated.";

	catalog<bool>("Debug Surface")=false;
	catalog<String>("Debug Surface","hint")=
			"Add lines to the output showing the surface refinement."
			"  These lines are put in a 'debug' group.";

	catalog<bool>("Debug Collisions")=false;
	catalog<String>("Debug Collisions","hint")=
			"Add lines to the output detailing the collisions."
			"  These lines are put in a 'debug' group.";

	catalog< sp<Component> >("Input Curves");
	catalog< sp<Component> >("Collider");

	catalog< sp<Component> >("Output Curves");
	catalog<String>("Output Curves","IO")="output";
	catalog<String>("Output Curves","copy")="Input Curves";
	catalog<bool>("Output Curves","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void BladeOp::handle(Record& a_rSignal)
{
#if FE_OBLD_DEBUG
	feLog("BladeOp::handle\n");
#endif

	//* access output curves (copy of input curves)
	m_spOutput=surfaceOutput(a_rSignal);
	if(m_spOutput.isNull())
	{
		feLog("BladeOp::handle"
				" output surface not SurfaceAccessibleI\n");
		catalog<String>("error")="inaccessible output";
		return;
	}

	//* access collider
	sp<SurfaceAccessibleI> spColliderAccessible=
			catalog< sp<Component> >("Collider");
	if(spColliderAccessible.isNull())
	{
		feLog("BladeOp::handle"
				" collider mesh not connected\n");
		catalog<String>("error")="missing collider";
		return;
	}

#if FALSE
	const U32 index=0;
	sp<SurfaceI> spCollider=spColliderAccessible->subSurface(index);
#else
	sp<SurfaceI> spCollider=spColliderAccessible->surface();
#endif
	if(spCollider.isNull())
	{
		feLog("BladeOp::handle"
				" collider input would not provide SurfaceI\n");
		catalog<String>("error")="incompatible collider";
		return;
	}

	const I32 refinement=catalog<I32>("Refinement");
	spCollider->setRefinement(refinement);

	sp<DrawI> spDrawI;
	const I32 threads=catalog<I32>("Threads");
	const BWORD debugSurface=catalog<bool>("Debug Surface");
	const BWORD debugCollisions=catalog<bool>("Debug Collisions");
	const BWORD recycle=catalog<bool>("Output Curves","recycle");
	const BWORD wantDebug=
			((debugSurface || debugCollisions) && threads==1);
	if(wantDebug && !recycle)
	{
		spDrawI=drawI(a_rSignal);
		if(!spDrawI.isValid())
		{
			feLog("BladeOp::handle DrawI invalid\n");
			catalog<String>("error")="invalid drawing interface";
			return;
		}
	}
	if(recycle==wantDebug)
	{
		catalog<bool>("Output Curves","recycle")=(!wantDebug);
	}

	m_blade.setDrawI(debugCollisions? spDrawI: sp<DrawI>(NULL));

	m_blade.setCollider(spCollider);

	String tangentName=catalog<String>("Tangent Attribute");
	String facingName=catalog<String>("Facing Attribute");

	m_blade.setTension(catalog<Real>("Tension"));
	m_blade.setRestoration(catalog<Real>("Restoration"));
	m_blade.setDrag(catalog<Real>("Drag"));
	m_blade.setBend(catalog<Real>("Bend"));
	m_blade.setGrabBias(catalog<Real>("Grab Bias"));
	m_blade.setCollisionGap(catalog<Real>("Collision Gap"));
	m_blade.setMass(catalog<Real>("Mass"));
	m_blade.setWind(catalog<Real>("Wind"));
	m_blade.setTurbulence(catalog<Real>("Turbulence"));
	m_blade.setCollisionMethod(
			Blade::CollisionMethod(catalog<I32>("Collision Method")));

	const Real timeStep=catalog<Real>("TimeStep");
	m_blade.setTimeStep(timeStep);

	static Real t=0.0;	// HACK
	t+=timeStep;
	m_blade.setTime(t);

	m_blade.populate();

	sp<SurfaceAccessorI> spAccessorTangent=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,tangentName);

	sp<SurfaceAccessorI> spAccessorFacing=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,facingName);

	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);

	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	std::vector<SpatialVector>& rTangent=m_blade.tangent();
	std::vector<SpatialVector>& rFacing=m_blade.facing();
	std::vector<SpatialVector>& rLocation=m_blade.location();
	std::vector<Real>& rVelocity=m_blade.velocity();

	//* inferred reset
	if(pointCount!=rLocation.size())
	{
		rLocation.resize(pointCount);
		for(U32 m=0;m<pointCount;m++)
		{
			rLocation[m]=spAccessorPoint->spatialVector(m,0);
		}

		const U32 pointCountV=subCount? pointCount*(subCount-1)/subCount: 0;
		rVelocity.resize(pointCountV);
		for(U32 m=0;m<pointCountV;m++)
		{
			rVelocity[m]=0.0;
		}

		rFacing.resize(pointCount);
		if(spAccessorFacing.isValid() &&
				spAccessorFacing->count() &&
				spAccessorFacing->subCount(0))
		{
			for(U32 m=0;m<pointCount;m++)
			{
				rFacing[m]=spAccessorFacing->spatialVector(m,0);
			}
		}
		else
		{
			feLog("BladeOp::handle facing attribute \"%s\" missing\n",
					facingName.c_str());

			for(U32 m=0;m<pointCount;m++)
			{
				set(rFacing[m]);
			}
		}

		const U32 pointCountT=subCount? pointCount/subCount: 0;
		rTangent.resize(pointCountT);
		if(spAccessorTangent.isValid() &&
				spAccessorTangent->count() &&
				spAccessorTangent->subCount(0))
		{
			for(U32 m=0;m<pointCountT;m++)
			{
				rTangent[m]=spAccessorTangent->spatialVector(m,0);
			}
		}
		else
		{
			feLog("BladeOp::handle tangent attribute \"%s\" missing\n",
					tangentName.c_str());

			for(U32 m=0;m<pointCountT;m++)
			{
				const Real angle=sin(0.7*m)+sin(0.3*m);

				set(rTangent[m],cos(angle),0.0,sin(angle));
			}
		}
		for(U32 m=0;m<pointCountT;m++)
		{
			m_blade.storeAsRest(m);
		}
	}

	if(spDrawI.isValid())
	{
		spDrawI->pushDrawMode(m_spDebugGroup);
	}

	setFullRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

	if(spDrawI.isValid())
	{
		if(debugSurface)
		{
			sp<DrawableI> spDrawable=spCollider;
			if(spDrawable.isValid())
			{
				spDrawable->draw(spDrawI,NULL);
			}
		}

		spDrawI->popDrawMode();
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::handle done\n");
#endif
}

void BladeOp::run(I32 a_id,Range a_range)
{
	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	String facingName=catalog<String>("Facing Attribute");
	sp<SurfaceAccessorI> spAccessorFacing=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,facingName);

	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);

	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	std::vector<SpatialVector>& rLocation=m_blade.location();
	std::vector<SpatialVector>& rFacing=m_blade.facing();

#if FE_OBLD_DEBUG
	feLog("BladeOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	for(U32 index=start;index<=end;index++)
	{
		m_blade.step(index);

		const U32 startSub=index*subCount;
		const U32 endSub=startSub+subCount;
		for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
		{
			spAccessorPoint->set(indexSub,0,rLocation[indexSub]);
		}
		if(spAccessorFacing.isValid() &&
				spAccessorFacing->count() &&
				spAccessorFacing->subCount(0))
		{
			for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
			{
				spAccessorFacing->set(indexSub,0,rFacing[indexSub]);
			}
		}
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::run done running thread %d\n",a_id);
#endif
}
