/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __plugin_Registry_h__
#define __plugin_Registry_h__

namespace fe
{

/**************************************************************************//**
	@brief Dynamic Library Manager

	@ingroup plugin

	Libraries can only be safely loaded and unloaded in a FILO
	(first-in last-out) manner.
	These restrictions are handled internally, but it may require
	some forethought to optimize memory usage by reducing library retention.
*//***************************************************************************/
class FE_DL_EXPORT Registry : public Handled<Registry>,
		public ObjectSafe<Registry>
{
	public:
						Registry(void);
						Registry(Master* pMaster);
virtual					~Registry(void);

		void			init(void);

						/** @brief Register the factories in the named
							dynamic library

							@em adaptname will append the platform-specific
							prefix and suffix to an abbreviated @em filename,
							such as from 'MyLib' to 'libMyLib.so',

							@em manageDependencies will automatically preload
							any libraries that the given library says
							it needs. */
		Result			manage(const String &filename, bool adaptname = true,
							bool manageDependencies = true);

						/** @brief (prototypical) Release interest in the
							named library

							This doesn't necessarily unload the library
							from memory.

							In the current system, trying to re-manage
							a library may be unpredictable. */
		Result			unmanage(const String &filename);

						/** @brief Potentially release libraries who don't
							have instantiated components */
		U32				prune(void);

						/** @brief Return a list of available implementations
							that match the string pattern */
		I32			listAvailable(const String &pattern,
								std::vector<String>& available);

						/** @brief Instantiate a Component of the given
							named implementation */
		sp<Component>	create(const String &implementation,
							BWORD quiet=FALSE) const;

						/// @brief Access the Master (ptr may not be valid)
		sp<Master>		master(void) const;

						/** @brief Access the component tracker

							@internal */
		Tracker&		tracker(void)					{ return m_tracker; };

const	String&			name(void) const				{ return m_name; }

		void			dump(void) const;
		void			dumpLibraries(void) const;

						/** @brief Instantiate a Counted of the given
							named implementation */
		sp<Counted>		create(String implementation,String name) const;

						/** @brief reference a Counted that is released
							just before unloading libraries

							Any number of dependents can be added. */
		void			addDependent(sp<Counted> spCounted)
						{	m_dependents.push_back(spCounted); }

	public:
	struct FactoryLocation
	{
		public:
							FactoryLocation(sp<Library> spLibrary,
									U32 factoryIndex):
									m_spLibrary(spLibrary),
									m_factoryIndex(factoryIndex)			{}
			sp<Library>		m_spLibrary;
			U32				m_factoryIndex;
	};
		std::map<String,FactoryLocation*>
						&factoryMap(void)				{ return m_factoryMap; }

	private:
		void			clear(void);

	class LoadedLibrary
	{
		public:
							LoadedLibrary(sp<Library> spLibrary):
									m_spLibrary(spLibrary),
									m_factoriesAccepted(0)
							{	FEASSERT(spLibrary.isValid()); }
							~LoadedLibrary(void);

			sp<Library>		m_spLibrary;
			U32				m_factoriesAccepted;
	};


		I32				findIndex(const String &filename);
		void			unmanage(U32 index);

		String			m_name;

		//*	Ordered FILO, one DL_Loader per dynamic library.
		std::vector<LoadedLibrary*>				m_loaderStack;

		std::vector< sp<Counted> >				m_dependents;

		std::map<String,FactoryLocation*>		m_factoryMap;

		Tracker									m_tracker;

		hp<Master>								m_hpMaster;
};


} /* namespace */

#endif /* __plugin_Registry_h__ */
